/*
 * File: intlist.c
 * Date:  09-03-99
 * Programmers:  Thomas Wu (thomas.wu@Stanford.Edu)
 *               Craig Nevill-Manning (nevill@cs.rutgers.edu)
 * Comment Writer:  Jimmy Y. Huang (yhuang@leland.stanford.edu)
 * ------------------------------------------------------------
 * This file implements the Intlist_T ADT.
 * ------------------------------------------------------------
 * Copyright 1999 The Board of Trustees of 
 * The Leland Stanford Junior University. All Rights Reserved.
 */


#include "intlist.h"
#include <stddef.h>
#include <stdlib.h>


/*
 * ----------
 * The statement #define T Intlist_T will replace all instances of the
 * character 'T' with the string "Intlist_T" after this point.
 * ----------
 */

#define T Intlist_T


/*
 * ----------
 * struct Intlist_T
 * ==========
 * The dynamic integer buffer exported by the Intlist_T ADT is implemented
 * using a linked list data structure.  The struct Intlist_T contains two
 * fields.  The int variable first is for storing the integer value that
 * corresponds to an integer in the original integer list.  The Intlist_T
 * variable rest, which is, again, a pointer-to-struct, points to
 * where in memory the next struct Intlist_T in the linked list of struct
 * Intlist_Ts is.  The viability of this linked list depends heavily on
 * the creation of the first link.  For the first struct Intlist_T the user
 * must call the function Intlist_push with the appropriate parameters as
 * illustrated in the example below:
 *
 *
 * Intlist_T namelist = (Intlist_T) NULL;
 *
 * c = fgetc (fp);
 * namelist = (Intlist_T) Intlist_push (c, namelist);
 *
 *
 * What the function Intlist_push will see are the character the stdio.h-
 * exported function fgetc has just obtained from the file stream and a
 * NULL pointer.  The function Intlist_push will see a NULL
 * pointer in the second parameter because the Intlist_T variable namelist,
 * a pointer-to-struct, has just been initialized to NULL.  The first
 * struct Intlist_T created will be the last one in the linked list of
 * struct Intlist_Ts (LIFO).
 * ----------
 */

struct T {
  int first;
  T rest;
};



/* Exported Functions */


/*
 * ----------
 * Intlist_push
 * ==========
 * The function Intlist_push creates a new struct Intlist_T and pushes
 * the newly created struct Intlist_T onto the front of the linked list
 * of struct Intlist_Ts.  The function Intlist_push has two parameters.
 * The first parameter is the int variable i, and the second parameter
 * is the Intlist_T pointer-to-struct list.  The int variable i stores
 * the integer that is to be added to the dynamic integer buffer.  If
 * the Intlist_T pointer-to-struct list is NULL, the integer stored in
 * the int variable i will be the first one added to the dynamic integer
 * buffer.  The Intlist_T pointer-to-struct that points to the longer
 * linked list will be returned to the caller.
 * ----------
 */

T Intlist_push (int i, T list)
{
  /*
   * ----------
   * The line of code below is the same as:
   *
   *      Intlist_T new = (Intlist_T) malloc (sizeof (struct Intlist_T));
   *
   * The struct Intlist_T contains two fields.  The int field first is
   * for storing the integer value that corresponds to an integer in the
   * original integer list.  The Intlist_T field rest, which is, again,
   * a pointer-to-struct, points to where in memory the next struct
   * Intlist_T in the linked list of struct Intlist_Ts is.
   * ----------
   */

  T new = (T) malloc (sizeof (*new));


  /*
   * ----------
   * The int field first in the newly created struct Intlist_T is set to
   * the value of the int variable i.
   * ----------
   */

  new -> first = i;


  /*
   * ----------
   * The Intlist_T pointer-to-struct field rest in the newly created struct
   * Intlist_T is set to where the second parameter Intlist_T pointer-to-
   * struct list is pointing in memory.  Where the second parameter list is
   * pointing is the current head of the linked struct Intlist_Ts.
   * ----------
   */

  new -> rest = list;


  /*
   * ----------
   * The location of the newly created struct Intlist_T is returned.
   * ----------
   */

  return (new);
}


/*
 * ----------
 * Intlist_free
 * ==========
 * When the function Intlist_free is called on a pointer to the caller's
 * Intlist_T pointer-to-struct, the linked list of the struct Intlist_Ts
 * to whose head the caller's Intlist_T pointer-to-struct intlist points
 * will be destroyed completely.
 * ----------
 */

void Intlist_free (T *intlist)
{
  T prev;


  /*
   * ----------
   * The function Intlist_free gets a copy of the address whereat the
   * Intlist_T pointer-to-struct intlist is located on the caller's
   * stack.  The value of the dereferenced intlist corresponds to where
   * in memory the caller's Intlist_T pointer-to-struct intlist is pointing.
   * The Intlist_T pointer-to-struct prev is then set to where the
   * dereferenced intlist is pointing.  Where both the dereferenced intlist
   * and the local Intlist_T pointer-to-struct prev are pointing to is the
   * first struct Intlist_T in the linked list of struct Intlist_Ts.  If
   * the Intlist_T pointer-to-struct prev is NULL, the linked list of struct
   * Intlist_Ts must have already been destroyed.
   * ----------
   */

  while (prev = *intlist) {


    /*
     * ----------
     * With the Intlist_T pointer-to-struct prev's pointing to the head of
     * the linked list that the Intlist_T pointer-to-struct intlist points,
     * the Intlist_T pointer-to-struct intlist can now be set to where
     * the second struct Intlist_T in the current linked list is.
     * ----------
     */

    *intlist = (*intlist) -> rest;


    /*
     * ----------
     * The block of memory the Intlist_T pointer-to-struct prev points to
     * can now be safely freed.  This block of memory used to be the first
     * struct Intlist_T in the linked list of struct Intlist_Ts.
     * ----------
     */

    free (prev);
  }  /* while (prev = *intlist) */
}


/*
 * ----------
 * Intlist_pop
 * ==========
 * The function Intlist_pop has two parameters.  The first parameter is
 * the Intlist_T pointer-to-struct list, and the second parameter is the
 * int pointer motif_index.  The function Intlist_pop takes the linked
 * list of struct Intlist_Ts, puts the integer value associated
 * with the first struct Intlist_T in the linked list into the
 * int variable the int pointer motif_index points to, and reclaims the
 * block of memory associated with the struct Intlist_T being popped.
 * The Intlist_T pointer-to-struct head now points to the head of the
 * shortened linked list, and it is returned to the caller.
 * ----------
 */

T Intlist_pop (T list, int *motif_index)
{
  T head;


  /*
   * ----------
   * Whereas the function Intlist_free gets a copy of the Intlist_T
   * pointer-to-struct intlist's memory address, the function Intlist_pop
   * gets a copy of the address to which the Intlist_T pointer-to-struct
   * list points.  The actual Intlist_T pointer-to-struct list on the
   * caller's stack and where it points to cannot be modified in anyway
   * by the function Intlist_pop, since all the function Intlist_pop has
   * is a copy of the memory address the caller's Intlist_T pointer-to-struct
   * list points to but not the memory address of the caller's Intlist_T
   * pointer-to-struct list.
   * ----------
   */

  if (list) {


    /*
     * ----------
     * The Intlist_T pointer-to-struct head is now initialized to where
     * the second struct Intlist_T in the linked list of
     * struct Intlist_Ts resides in memory..
     * ----------
     */

    head = list -> rest;


    /*
     * ----------
     * The int variable on the caller's stack the pointer motif_index is
     * pointing to is set to the integer value associated with the first
     * struct Intlist_T in the linked list of struct Intlist_Ts.
     * ----------
     */

    *motif_index = list -> first;


    /*
     * ----------
     * The memory associated with the first struct Intlist_T in the linked
     * list of struct Intlist_Ts is reclaimed as the local copy of the
     * Intlist_T pointer-to-struct list is passed into the free function.
     * The caller's copy of the Intlist_T pointer-to-struct list, on the
     * other hand, is still pointing to the struct Intlist_T that used
     * to be the head of the linked list of struct Intlist_Ts.
     * ----------
     */

    free (list);


    /*
     * ----------
     * The pointer to the new first struct Intlist_T in the linked list of
     * struct Intlist_Ts (head) is returned to the caller.  Because the
     * caller's copy of the Intlist_T pointer-to-struct list is still
     * pointing to where the old first struct Intlist_T used to be, the
     * caller's copy of the Intlist_T pointer-to-struct list must be
     * reset to where the Intlist_T pointer-to-struct is pointing.  An
     * example is shown below:
     *
     *      seqlist = Intlist_pop (seqlist, &c);
     *
     * If the Intlist pointer-to-struct seqlist is not re-assigned to
     * the Intlist pointer-to-struct returned by Intlist_pop function,
     * it will still be pointing to a memory location that has been
     * reclaimed by the system (Yes, very bad).
     * ----------
     */

    return (head);
  } else /* if (list) */ {


    /*
     * ----------
     * The linked list of struct Intlist_Ts must be empty, because the
     * Intlist_T pointer-to-struct list is NULL.  The caller's copy of
     * the Intlist_T pointer-to-struct list will be set to NULL.
     * ----------
     */

    return (list);
  }  /* if (list) */
}


