/*
 * File:  sequence.c
 * Date:  09-07-99
 * Programmers:  Thomas Wu (thomas.wu@Stanford.Edu)
 *               Craig Nevill-Manning (nevill@cs.rutgers.edu)
 * Comment Writer:  Jimmy Y. Huang (yhuang@leland.stanford.edu)
 * ------------------------------------------------------------
 * This file implements the Sequence_read function the sequence.h
 * file exports.
 * ------------------------------------------------------------
 * Copyright 1999 The Board of Trustees of 
 * The Leland Stanford Junior University. All Rights Reserved.
 */


#include <ctype.h>
#include "intlist.h"
#include "sequence.h"
#include <stdlib.h>
#include <string.h>


/*
 * ----------
 * Private Function Prototypes
 * ==========
 * Tom's version of this code does not include the function declarations
 * below, because he does not like to "write things twice."  By omitting
 * function declarations, he restricts his function definitions to his
 * original arrangement, so functions that call on other functions are
 * always placed below the functions they call on.
 * ----------
 */

static char *Sequence_read_name (FILE *fp);
static char *Sequence_read_seq (FILE *fp, int *length, char format);


/*
 * ----------
 * Sequence_read_name
 * ==========
 * The private function Sequence_read_name takes the FILE pointer fp as
 * an argument.  The function Sequence_read_name returns a char pointer
 * to a dynamic array containing the name of the one sequence being
 * read from the input file the FILE pointer fp points to.  The function
 * Sequence_read_name returns NULL if no name has been read.
 * ----------
 */

static char *Sequence_read_name (FILE *fp)
{
  char *name_char;


  /*
   * ----------
   * The int variable n is a counter for the number of integers in the
   * dynamic integer buffer the file intlist.h exports.  The int variable
   * i is an index variable for the dynamic array the char pointer
   * name_char points to.
   * ----------
   */

  int n = 0, i, c;
  Intlist_T namelist = (Intlist_T) NULL;


  /*
   * ----------
   * The following while loop is responsible for reading the name of one
   * FASTA-formatted input sequence and putting the ASCII value of each
   * character from that name in the dynamic integer buffer the file
   * intlist.h exports.  An example of a FASTA-formatted input sequence is
   * shown below:
   *
   * >THTM_HUMAN\n
   * MASPQLCRAL VSAQWVAEAL RAPRAGQPLQ LLDASWYLPK LGRDATQFEE RHIPGAAFFD\n
   * IDQCSDRTSP YDHMLPGAEH FAEYAGRLGV GAATHVVIYD ASDQGLYSAP RVWWMFRAFG\n
   * HHAVSLLDGG LRHWLRQNLP LSSGKSQPAP AEFRAQLDPA FIKTYEDIKE NLESRRFQVV\n
   * DSRATGRFRG TEPEPRDGIE PGHIPGTVNI PFTDFLSQEG LEKSPEEIRH LFQEKKVDLS\n
   * KPLVATCGSG VTACHVALGA YLCGKPDVPI YDGSWVEWYM RARPEDVISE GRGKTH\n
   *
   * As the above example illustrates, the name of the input sequence
   * begins with a '>' character and ends with a newline character '\n'.
   * The actual input sequence comes after that newline character '\n'
   * immediately.
   * ----------
   */

  while (!feof (fp) && ( (c = fgetc (fp)) != '\n' )) {


    /*
     * ----------
     * For the very first ASCII value to be pushed onto the dynamic integer
     * buffer the file intlist.h exports, the function Intlist_push will see
     * a NULL pointer in the second parameter because the Intlist_T variable
     * namelist, a pointer-to-struct, has just been initialized to NULL
     * before the program stepped into the while loop.  The first
     * ASCII value pushed onto the dynamic integer buffer will be the last
     * ASCII value popped out (LIFO).  The counting variable n keeps track
     * of the number of ASCII values in the intlist.h-exported dynamic
     * integer buffer.
     * ----------
     */

    namelist = (Intlist_T) Intlist_push (c, namelist);
    ++n;
  }  /* while (!feof (fp) && ((c = fgetc (fp)) != '\n')) */


  /*
   * ----------
   * If this private function Sequence_read_name is called, the end-of-file
   * character should not be encountered during the execution of this
   * private function.  If an end-of-file character is encountered, the
   * input sequence file must have been truncated somehow and should be fixed.
   * ----------
   */

  if (feof (fp)) {
    fprintf (stderr, "File ended while reading name of sequence.\n");


    /*
     * ----------
     * Tom originally wrote:
     *
     * return ((char *) NULL);
     *
     * If the input sequence file prematurely terminates while the name
     * of a FASTA-formatted sequence is being read, the program really
     * should exit, so the user may fix the problem immediately.
     * ----------
     */

    exit (1);
  }  /* if (feof (fp)) */


  /*
   * ----------
   * If the int variable n is 0, the Intlist_T pointer-to-struct namelist
   * must still be NULL.  No name, therefore, has been read.  The NULL
   * pointer is returned to the calling function.
   * ----------
   */

  if (n == 0) {
    return ((char *) NULL);
  } else /* if (n == 0) */ {


    /*
     * ----------
     * If the int variable n is non-zero, at least one ASCII value must
     * have been pushed onto the dynamic integer buffer the file intlist.h
     * exports.  The content of that dynamic integer buffer must be
     * transfered to a dynamic char array before this function exits.
     * An array of (n + 1) char elements is created on the heap.  The extra
     * char element is to house the NULL character '\0'.
     * ----------
     */

    name_char = (char *) calloc (n + 1, sizeof (char));

    name_char[n] = '\0';


    /*
     * ----------
     * The following for loop is responsible for putting ASCII values stored
     * in the dynamic integer buffer the file intlist.h exports into a
     * dynamic char array in the correct order.  The integers in the
     * intlist.h-exported dynamic buffer are stored in reverse order; that
     * is, the first integer "popped" out of the dynamic integer buffer is
     * always going to be the integer most recently read (LIFO).  To put the
     * ASCII values in the intlist.h-exported dynamic buffer into the correct
     * order, the first ASCII value popped out and assigned to the char
     * variable c must go into the (n - 1)th char element in the dynamic
     * array the char pointer name_char points to; the last ASCII value
     * popped out and assigned to the char variable c, on the other hand,
     * must go into the 0th char element in the dynamic array the char
     * pointer name_char points to.
     * ----------
     */

    for (i = n - 1; i >= 0; --i) {
      namelist = Intlist_pop (namelist, &c);
      name_char[i] = (char) c;
    }  /* for (i = n - 1; i >= 0; --i) */


    /*
     * ----------
     * The char pointer name_char, which points to the dynamic array
     * that contains the name of the FASTA-formatted input sequence
     * about to be read, is returned to the Sequence_read function.
     * ----------
     */

    return (name_char);
  }  /* if (n == 0) */
}


/*
 * ----------
 * Sequence_read_seq
 * ==========
 * The private function Sequence_read_seq takes the FILE pointer fp, a
 * pointer to the user-supplied int variable length, and the char
 * variable format as arguments.  The char variable format indicates to
 * the function Sequence_read_seq which of the two formats, FASTA or
 * sequence-by-sequence, the input sequence file is in.  The function
 * Sequence_read_seq returns a char pointer to a dynamic array containing
 * one sequence from the input file the FILE pointer fp points to.  The
 * function Sequence_read_seq returns NULL if no sequence has been read.
 * ----------
 */

static char *Sequence_read_seq (FILE *fp, int *length, char format)
{
  char *sequence_char;


  /*
   * ----------
   * The int variable n is a counter for the number of ASCII values in the
   * dynamic integer buffer the file intlist.h exports.  The int variable
   * i is an index variable for the dynamic array the char pointer
   * sequence_char points to.
   * ----------
   */

  int n = 0, i, c;
  Intlist_T seqlist = (Intlist_T) NULL;


  /*
   * ----------
   * The following while loop is responsible for reading one input
   * sequence in either the sequence-by-sequence format or the FASTA
   * format.  An example of a FASTA-formatted input sequence is
   * shown below:
   *
   * >THTM_HUMAN\n
   * MASPQLCRAL VSAQWVAEAL RAPRAGQPLQ LLDASWYLPK LGRDATQFEE RHIPGAAFFD\n
   * IDQCSDRTSP YDHMLPGAEH FAEYAGRLGV GAATHVVIYD ASDQGLYSAP RVWWMFRAFG\n
   * HHAVSLLDGG LRHWLRQNLP LSSGKSQPAP AEFRAQLDPA FIKTYEDIKE NLESRRFQVV\n
   * DSRATGRFRG TEPEPRDGIE PGHIPGTVNI PFTDFLSQEG LEKSPEEIRH LFQEKKVDLS\n
   * KPLVATCGSG VTACHVALGA YLCGKPDVPI YDGSWVEWYM RARPEDVISE GRGKTH\n
   *
   * As the above example illustrates, the name of the input sequence
   * begins with a '>' character and ends with a newline character '\n'.
   * The actual input sequence comes after that newline character '\n'
   * immediately.  An example of one input sequence found in the
   * sequence-by-sequence format is shown below:
   *
   * MASPQLCRAL VSAQWVAEAL RAPRAGQPLQ LLDASWYLPK LGRDATQFEE RHIPGAAFFD
   * IDQCSDRTSP YDHMLPGAEH FAEYAGRLGV GAATHVVIYD ASDQGLYSAP RVWWMFRAFG
   * HHAVSLLDGG LRHWLRQNLP LSSGKSQPAP AEFRAQLDPA FIKTYEDIKE NLESRRFQVV
   * DSRATGRFRG TEPEPRDGIE PGHIPGTVNI PFTDFLSQEG LEKSPEEIRH LFQEKKVDLS
   * KPLVATCGSG VTACHVALGA YLCGKPDVPI YDGSWVEWYM RARPEDVISE GRGKTH\n
   *
   * The example above has only 1 newline character '\n'.  All the other
   * apparent line breaks are put in, so the entire sequence can fit on
   * one page.  The program assumes an input sequence has come to an end
   * once a newline character '\n' has been encountered.
   * ----------
   */

  while (!feof (fp) && ((c = fgetc (fp)) != format)) {


    /*
     * ----------
     * If the character just read is not an alphabet as determined by the
     * isalpha function exported by the file ctype.h, it is skipped.
     * ----------
     */

    if (isalpha (c)) {


      /*
       * ----------
       * For the very first ASCII value to be pushed onto the dynamic integer
       * buffer the file intlist.h exports, the function Intlist_push will
       * see a NULL pointer in the second parameter because the Intlist_T
       * variable seqlist, a pointer-to-struct, has just been initialized to
       * NULL before the program stepped into the while loop.  The first
       * ASCII value pushed onto the dynamic integer buffer will be the last
       * ASCII value popped out (LIFO).  The counting variable n keeps track
       * of the number of ASCII values in the intlist.h-exported dynamic
       * integer buffer.
       * ----------
       */

      seqlist = (Intlist_T) Intlist_push (c, seqlist);
      ++n;
    }  /* if (isalpha (c)) */
  }  /* while (!feof (fp) && ((c = fgetc (fp)) != format)) */


  if (c == '>') {
    ungetc(c,fp);
  }


  /*
   * ----------
   * The int variable length on the calling function's stack is now set to
   * the length of the input sequence.  If the int variable n is now 0, the
   * Intlist_T pointer-to-struct seqlist must still be NULL.  No sequence,
   * therefore, has been read.  The NULL pointer is returned to the calling
   * function.
   * ----------
   */

  *length = n;

  if (n == 0) {
    return ((char *) NULL);
  } else /* if (n == 0) */ {


    /*
     * ----------
     * If the int variable n is non-zero, at least one ASCII value must
     * have been pushed onto the dynamic integer buffer the file intlist.h
     * exports.  The content of that dynamic integer buffer must be
     * transfered to a dynamic char array before this function exits.
     * An array of (n + 1) char elements is created on the heap.  The extra
     * char element is to house the NULL character '\0'.
     *
     * Tom originally wrote:
     *
     * sequence_char = (char *) calloc (n, sizeof (char));
     *
     * He did not make room for a NULL character, because he never had to use
     * sequence_char[n] and never treated sequence_char as a string in
     * emotif-3.0/src/emotif-search/emotif-search.c.  Tom's original code
     * puts an unnecessary constraint on the clients of sequence.h.
     * ----------
     */

    sequence_char = (char *) calloc (n + 1, sizeof (char));
    sequence_char[n] = '\0';


    /*
     * ----------
     * The following for loop is responsible for putting ASCII values stored
     * in the dynamic integer buffer the file intlist.h exports into a
     * dynamic char array in the correct order.  The integers in the
     * intlist.h-exported dynamic buffer are stored in reverse order; that
     * is, the first integer "popped" out of the dynamic integer buffer is
     * always going to be the integer most recently read (LIFO).  To put the
     * ASCII values in the intlist.h-exported dynamic buffer into the correct
     * order, the first ASCII value popped out and stored in the char
     * variable c must go into the (n - 1)th char element in the dynamic
     * array the char pointer sequence_char points to; the last ASCII value
     * popped out and assigned to the char variable c, on the other hand,
     * must go into the 0th char element in the dynamic array the char
     * pointer name_char points to.
     * ----------
     */

    for (i = n - 1; i >= 0; --i) {
      seqlist = Intlist_pop (seqlist, &c);
      sequence_char[i] = (char) c;
    }  /* for (i = n - 1; i >= 0; --i) */


    /*
     * ----------
     * The char pointer sequence_char, which points to the dynamic array
     * that contains the actual input sequence, is returned to the
     * Sequence_read function.
     * ----------
     */

    return (sequence_char);
  }  /* if (*length == 0) */
}


/*
 * ----------
 * Sequence_read
 * ==========
 * The exported function Sequence_read takes three arguments.  The first
 * argument is a pointer to the file containing the input sequences.  The
 * second argument is a pointer to the char pointer name; the char pointer
 * name, after the private function Sequence_read_name is called, will point
 * to a dynamic array containing the name of one input sequence.  The third
 * argument is a pointer to an int variable length; this int variable length,
 * after the private function Sequence_read_seq is called, will contain the
 * length of the input sequence just read.  The function Sequence_read
 * returns a char pointer to a dynamic array, created by the private function
 * Sequence_read_seq.  This dynamic char array contains one input sequence
 * from the user-supplied file.  The memory de-allocation responsibilities
 * lie with the users of this library.  The function Sequence_read returns
 * NULL if no sequence has been read.
 * ----------
 */

char *Sequence_read (FILE *fp, char **name, int *length)
{
  char c;

  if (feof (fp)) {
    return (char *) NULL;
  } else /* if (feof (fp)) */ {


    /* 
     * ----------
     * The input file is parsed character by character using the stdio.h-
     * exported function fgetc.  If the first character read from the FILE
     * pointer fp is '>', the one input sequence about to be read is assumed
     * to be in the FASTA format.
     * ----------
     */

    if ((c = fgetc (fp)) == '>') {


      /*
       * ----------
       * The result from the private function Sequence_read_name is the
       * memory location of the dynamic char array containing the
       * name of the input sequence.  The user-supplied char **name is
       * dereferenced once, so char *name on the calling function's
       * stack may be set to where that dynamic char array is.
       * ----------
       */

      *name = Sequence_read_name (fp);


      /*
       * ----------
       * The result from the private function Sequence_read_seq is the
       * memory location of the dynamic char array containing the actual
       * input sequence whose name has just been obtained by the private
       * function Sequence_read_name.  The memory location is returned
       * directly to the user.  If no sequence has been read, the function
       * Sequence_read_seq returns NULL.
       * ----------
       */

      return (Sequence_read_seq (fp, length, '>'));
    } else /* if ((c = fgetc (fp)) == '>') */ {


    /* 
     * ----------
     * If the first character read from the FILE pointer fp is not '>', the
     * one input sequence about to be read is assumed to be in the
     * sequence-by-sequence format.  That first character must be placed
     * back in the file stream using the stdio.h-exported function ungetc.
     * Since the one input sequence about to be read has already been
     * determined to be of the sequence-by-sequence format, the user-
     * supplied char **name may now be dereferenced once, so the char *name
     * on the calling function's stack may be set to NULL.
     * ----------
     */

      ungetc (c, fp);
      *name = (char *) NULL;


      /*
       * ----------
       * The result from the private function Sequence_read_seq is the
       * memory location of the dynamic char array containing the actual
       * input sequence whose name has just been obtained by the private
       * function Sequence_read_name.  The memory location is returned
       * directly to the user.  If no sequence has been read, the function
       * Sequence_read_seq returns NULL.
       * ----------
       */

      return (Sequence_read_seq (fp, length, '\n'));
    }  /* if ((c = fgetc (fp)) == '>') */
  }  /* if (feof (fp)) */
}







