/*
 * File: reformat-fasta.c
 * Date:  09-09-99
 * Programmers:  Thomas Wu (thomas.wu@Stanford.Edu)
 *               Craig Nevill-Manning (nevill@cs.rutgers.edu)
 * Comment Writer:  Jimmy Y. Huang (yhuang@leland.stanford.edu)
 * ------------------------------------------------------------
 * The program reformat-fasta converts a FASTA-formatted input
 * sequence file into a format the emotif-3.0/cgi-bin/nph-emotif-scan
 * can process.  For emotif-2.0 this program was implemented using
 * Perl.
 * ------------------------------------------------------------
 * Copyright 1999 The Board of Trustees of 
 * The Leland Stanford Junior University. All Rights Reserved.
 */


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>


int main (int argc, char *argv[]) {
  FILE *fasta_fp;
  char *fasta_file;
  char c;
  

  /*
   * ----------
   * The variable argc of type int is the argument count.  The
   * statement
   *
   *      reformat-fasta pdb.fasta
   *
   * is broken up into 2 parts:  argv[0] = "reformat-fasta" and
   * argv[1] = "pdb.fasta".  The value of argc, therefore, is 2.  The
   * program expects a FASTA-formatted sequence file.  If the stdio.h-
   * exported fopen function cannot open the file whose name the char
   * pointer fasta_file points to, the fopen function returns NULL, and
   * the FILE pointer fasta_fp is, as a consequence, set to NULL.
   * ----------
   */

  fasta_file = argv[1];
  fasta_fp = fopen (fasta_file, "r");

  if (!fasta_fp) {
    fprintf (stderr,"Can't open file $s\n", fasta_file);
    exit(9);
  }


  /*
   * ----------
   * The program reformat-fasta assumes that anything it encounters before
   * the first '>' character is junk.
   * ----------
   */

  while (!feof(fasta_fp) && ((c = fgetc(fasta_fp)) != '>')) ;


  /*
   * ----------
   * The program reformat-fasta will remain in the following while loop
   * as long as there are more characters in the FASTA-formatted input
   * sequence file to be read.
   * ----------
   */

  while (!feof (fasta_fp)) {


    /*
     * ----------
     * At this point of the program, a '>' character has just been read.
     * Any character encountered from now on before the reading of a newline
     * character '\n' is going to be part of the name or description for
     * the one input sequence about to be processed.
     * ----------
     */

    while (!feof (fasta_fp) && ((c = fgetc(fasta_fp)) != '\n')) {


      /*
       * ----------
       * Every character in the name or description for the one input
       * sequence about to be processed is fed to standard output using
       * the stdio.h-exported putchar function.  The putchar function is
       * equivalent to putc (c, stdout), another stdio.h-exported function.
       * ----------
       */

      putchar(c);
    }


    /*
     * ----------
     * At this point of the program, a newline character '\n' has just been
     * encountered.  Any character encountered from now on before the reading
     * of a '>' character is going to be part of the input sequence.  A tab
     * character '\t' is fed to the standard output to separate the name or
     * description of the sequence about to be read from the sequence itself. 
     * ----------
     */

    putchar ('\t');

    while (!feof(fasta_fp) && ((c = fgetc(fasta_fp)) != '>')) {


      /*
       * ----------
       * If the char variable c is part of the English alphabet, the program
       * will print out the char variable c in lower case to the standard
       * output.
       *
       * Tom originally wrote:
       *
       * if (!isspace (c)) {
       *   putchar (tolower (c));
       * }
       *
       * 
       * ----------
       */

      if (isalpha (c)) {
	putchar (tolower (c));
      }
    }


    /*
     * ----------
     * At this point of the program a '>' character has just been
     * encountered, and the entire input sequence should have all been
     * fed to the standard output.  A newline character '\n' is inserted
     * to mark the end of one input sequence.
     * ----------
     */

    putchar ('\n');
  }

  fclose (fasta_fp);

  return (0);
}

